from io import BytesIO
import matplotlib.pyplot as plt
from IPython.display import HTML
import pandas as pd


class Artifact:
    """Displays or saves an artifact."""

    extension = ""

    def __init__(self, name, file):
        self.name = name
        self.file = file
        self._content = None

    def __repr__(self):
        return f'{self.__class__.__name__}(name={self.name})'

    def save(self):
        with open(self._make_filename(), 'wb') as file:
            file.write(self.content)

    @property
    def content(self):
        if self._content is None:
            self._content = self.file.read()
        return self._content

    def _make_filename(self):
        parts = self.file.filename.split('/')
        return f'{parts[-2]}_{parts[-1]}.{self.extension}'


class PNGArtifact(Artifact):
    """Displays or saves a PNG artifact."""

    extension = "png"

    def __init__(self, name, file):
        super().__init__(name, file)
        self._img = None

    def show(self, figsize=(10, 10)):
        fig, ax = plt.subplots(figsize=figsize)
        ax.imshow(self.img)
        ax.axis('off')
        return fig

    @property
    def img(self):
        if self._img is None:
            self._img = plt.imread(BytesIO(self.content))
        return self._img


class MP4Artifact(Artifact):
    """Displays or saves a MP4 artifact"""

    extension = "mp4"

    def __init__(self, name, file):
        super().__init__(name, file)
        self.movie = None

    def show(self):
        if self.movie is None:
            self._make_movie()
        return self.movie

    def _make_movie(self):
        self.save()
        self.movie = HTML(f"""
        <video width="640" height="480" controls autoplay>
          <source src="{self._make_filename()}" type="video/mp4">
        </video>
        """)


class CSVArtifact(Artifact):
    """Displays and saves a CSV artifact"""

    extension = "csv"

    def __init__(self, name, file):
        super().__init__(name, file)
        self.df = None

    def show(self):
        if self.df is None:
            self.df = self._make_df()
        return self.df

    def _make_df(self):
        df = pd.read_csv(self.file)
        return df



from functools import reduce
from operator import and_, or_
from incense.experiment import Experiment
from typing import *


def find_differing_config_keys(experiments: Iterable[Experiment]):
    """Find the config keys that were assigned to different values in a cohort of experiments.."""
    config_values = []
    configs = []
    for experiment in experiments:
        config_values.append(set(str(v) for v in experiment.config.values()))
        configs.append(experiment.config)

    differing_config_values = reduce(or_, config_values) ^ reduce(and_, config_values)
    differing_config_keys = set()
    for config in configs:
        for key, value in config.items():
            if str(value) in differing_config_values:
                differing_config_keys.add(key)

    return differing_config_keys


def format_config(config, *config_keys):
    return ' | '.join(f'{key}={config[key]}' for key in config_keys)



from pymongo import MongoClient
import gridfs
from functools import lru_cache
from typing import *

from incense.experiment import Experiment

credentials = []

MAX_CACHE_SIZE = 32

class ExperimentLoader:
    """Loads artifacts related to experiments."""

    def __init__(self, mongo_uri, db_name):
        client = MongoClient(mongo_uri)
        self._database = client[db_name]
        self._runs = self._database.runs
        self._grid_filesystem = gridfs.GridFS(self._database)

    def find_by_ids(self, experiment_ids: Iterable[int]) -> List[Experiment]:
        """
        Find experiments based on a collection of ids.

        Args:
            experiment_ids: Iterable of experiment ids.

        Returns:
            The experiments corresponding to the ids.
        """
        experiments = [self.find_by_id(experiment_id) for experiment_id in experiment_ids]

        return experiments

    # The cache makes sure that retrieval of the experiments
    # is not unnecessarily done more than once.
    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_by_id(self, experiment_id: int) -> Experiment:
        """
        Find experiment based on its id.

        Args:
            experiment_id: The id  of the experiment.

        Returns:
            The experiment corresponding to the id.
        """
        experiment = self._find_experiment(experiment_id)

        return self._make_experiment(experiment)

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_by_name(self, name: str) -> List[Experiment]:
        """
        Find experiments based on regex search against its name.

        A partial match between experiment name and regex is enough
        to find the experiment.

        Args:
            name: Regex that is matched against the experiment name.

        Returns:
            The matched experiments.
        """
        return self.find_by_key('experiment.name', name)

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_by_config_key(self, key: str, value: str) -> List[Experiment]:
        """
        Find experiments based on regex search against a configuration value.

        A partial match between configuration value and regex is enough
        to find the experiment.

        Args:
            key: Configuration key to search on.
            value: Regex that is matched against the experiment's configuration.

        Returns:
            The matched experiments.
        """
        cursor = self._runs.find({f'config.{key}': {'$regex': rf'{value}'}})
        experiments = [self._make_experiment(experiment) for experiment in cursor]
        return experiments

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_by_key(self, key: str, value: str) -> List[Experiment]:
        """
        Find experiments based on regex search against a value stored in the database.

        A partial match between the value and the regex is enough
        to find the experiment.

        Args:
            key: Key to search on.
            value: Regex that is matched against the experiment's configuration.

        Returns:
            The matched experiments.
        """
        cursor = self._runs.find({key: {'$regex': rf'{value}'}})
        experiments = [self._make_experiment(experiment) for experiment in cursor]
        return experiments

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def _find_experiment(self, experiment_id: int):
        return self._runs.find_one({'_id': experiment_id})

    def _make_experiment(self, experiment) -> Experiment:
        return Experiment.from_db_object(self._database, self._grid_filesystem, experiment)



import pandas as pd
from typing import *

from incense import artifact
from incense.artifact import Artifact

content_type_to_artifact_cls = {
    'image/png': artifact.PNGArtifact,
    'text/csv': artifact.CSVArtifact
}


class Experiment:

    def __init__(self, id_, database, grid_filesystem, config, artifact_links):
        self.id = id_
        self.config = config
        self._artifacts_links = artifact_links
        self._database = database
        self._grid_filesystem = grid_filesystem
        self._artifacts = None
        self._metrics = None

    def __repr__(self):
        return f'{self.__class__.__name__}(id={self.id})'

    @classmethod
    def from_db_object(cls, database, grid_filesystem, experiment_data: dict):
        config = experiment_data['config']
        artifacts_links = experiment_data['artifacts']
        id_ = experiment_data['_id']
        return cls(id_, database, grid_filesystem, config, artifacts_links)

    @property
    def artifacts(self) -> Dict[str, artifact.Artifact]:
        """
        The artifacts belonging to the experiment.

        Returns:
            A mapping from artifact names to artifact objects, that
            belong to the experiment.
        """
        if self._artifacts is None:
            self._artifacts = self._load_artifacts()

        return self._artifacts

    @property
    def metrics(self) -> Dict[str, pd.Series]:
        """
        The metrics belonging to the experiment.

        Returns:
            A mapping from metric names to pandas Series objects, that
            belong to the experiment.
        """
        if self._metrics is None:
            self._metrics = self._load_metrics()

        return self._metrics

    def _load_artifacts(self) -> Dict[str, Artifact]:
        artifacts = {}
        for artifact_link in self._artifacts_links:
            artifact_file = self._grid_filesystem.get(artifact_link['file_id'])
            artifact_type = content_type_to_artifact_cls[artifact_file.metadata['content-type']]
            name = artifact_link['name']
            artifacts[name] = artifact_type(name, artifact_file)
        return artifacts

    def _load_metrics(self) -> Dict[str, pd.Series]:
        metrics = {}
        metric_db_entries = self._database.metrics.find({'run_id': self.id})
        for metric_db_entry in metric_db_entries:
            metrics[metric_db_entry['name']] = pd.Series(data=metric_db_entry['values'],
                                                         index=pd.Index(metric_db_entry['steps'], name='step'),
                                                         name=metric_db_entry['name'])
        return metrics



from setuptools import setup
import os

_here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='incense',
    version='0.0.1',
    description='Interactively retrieve data from sacred experiments.',
    long_description=long_description,
    author='Rüdiger Busche',
    author_email='rbusche@uos.de',
    url='https://github.com/JarnoRFB/incense',
    license='MIT',
    packages=['incense'],
    install_requires=[
        'sacred',
        'matplotlib>=3',
        'pandas>=0.23',
        'jupyterlab>=0.35',
        'pymongo>=3.7'
    ],
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
)
